/* NAME:                                
 * 	daemon - make caller a daemon
 * 
 * SYNOPSIS:
 *	int daemon(int nochdir, int noclose);
 * 	pid_t pid = daemon_init(int ttyf, char *name);
 * 
 * DESCRIPTION:
 *	Daemon() is equivalent to the 4.4BSD function of the same
 *	name, but handles non-BSD systems.  If "nochdir" is 0 it will
 *	chdir to ``/''.  If "noclose" is 0, it will associate stdio
 *	descriptors with '/dev/null'.
 *
 *	Daemon_init() uses daemon() to dissociate the caller from its
 *	controlling terminal and if "ttyf" is not 0, check up to
 *	MAX_FD's for open tty devices.   If "ttyf" is 1, open tty
 *	descriptors are closed and then re-opened, otherwise they are
 *	associated with '/dev/null'. 
 *      
 *      If "name" is not NULL, it is assumed to be the name of a 
 *      of the daemon process.  A call is made to LogMyPID() to 
 *      record the daemon's PID in a file named "name.pid" in 
 *      the directory named by LOGPID_DIR or the current directory.
 *      
 * RETURNS:
 *      Returns pid or -1 after dissociating from tty devices.
 *      If compiled with -DEXIT_ON_ERR,  exits if something goes 
 *      wrong. 
 *
 * AUTHOR:
 *	Simon J. Gerraty <sjg@zen.void.oz.au>
 */
/*
 *      @(#)Copyright (c) 1993, Simon J. Gerraty.
 *      
 *      This is free software.  It comes with NO WARRANTY.
 *      Permission to use, modify and distribute this source code 
 *      is granted subject to the following conditions.
 *      1/ that the above copyright notice and this notice 
 *      are preserved in all copies and that due credit be given 
 *      to the author.  
 *      2/ that any changes to this code are clearly commented 
 *      as such so that the author does not get blamed for bugs 
 *      other than his own.
 *      
 *      Please send copies of changes and bug-fixes to:
 *      sjg@zen.void.oz.au
 */
#ifndef lint
static char *RCSid = "$Id: daemon.c,v 1.4 1995/05/08 13:10:38 sjg Exp $";
#endif
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#ifndef HAVE_DAEMON
#include <stdio.h>
#if defined(__STDC__)
# include <stdlib.h>
#endif
#include <string.h>
#include <sys/types.h>
#ifdef HAVE_SYS_STDTYPES_H
# include <sys/stdtypes.h>
#endif
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/cdefs.h>

#undef daemon	

#if defined(sun386)
typedef int pid_t;
#endif

extern char *getenv __P((const char *));
extern char *ttyname  __P((int));
  
#ifndef MAXFILENAME
# define MAXFILENAME 64
#endif

/*
 * we check this many file descriptors for tty devices
 */
#ifndef MAX_FD
# define MAX_FD 5
#endif
/*
 * we try the fork() this many times
 */
#ifndef MAX_TRY
# define MAX_TRY 5
#endif

/*
 * pick which you like, and update the manual page
 */
#ifdef EXIT_ON_ERR
# define ERRET exit(1)
#else
# define ERRET return (-1)
#endif

/*
 * we look for this name in the environment
 * to tell us where to put "name.pid"
 */
#ifndef LOGPID_ENV
# define LOGPID_ENV "PID_DIR"
#endif
/*
 * otherwise we put it here
 */
#ifndef LOGPID_DIR
# define LOGPID_DIR "."
#endif

#ifndef HAVE_DAEMON
/*
 * this needs to be compatible with the 4.4BSD version,
 * but work on other systems too...
 */

int
daemon(nochdir, noclose)
	int nochdir, noclose;
{
	pid_t pid;
	int i;

#ifndef HAVE_SETSID
	setpgrp(0, 0);
#endif
	/*
	 * fork(2) can fail due to a temp resource limit,
	 * try a few times before giving up...
	 */
	for (i = 0; i < MAX_TRY; i++) {
		if ((pid = fork()) >= 0)
			break;
		sleep(i + 1);			/* sleep longer each re-try */
	}
	switch (pid) {
	case -1:
		perror("fork");
		ERRET;				/* NOT REACHED */
	case 0:					/* child */
#ifdef HAVE_SETSID
		(void) setsid();		/* if it exists, we must call it */
#endif
		if (!nochdir)
			(void) chdir("/");

		if (!noclose) {
			int devnull = open("/dev/null", O_RDWR, 0);

			if (devnull != -1) {
				for (i = 0; i < 3; i++) {
					(void) close(i);
					(void) dup(devnull);
				}
				if (devnull > 2)
					(void) close(devnull);
			}
		}
		break;
	default:				/* parent */
		exit(0);			/* NOT REACHED */
	}
	return 0;
}
#endif



/*
 * log our PID in a file
 */

static int 
LogMyPID (pid, name)
	pid_t pid;
	char *name;
{
  FILE *fh;
  char logF[MAXFILENAME];

  if (name)				/* just in case */
  {
    (void) sprintf(logF, "%s.pid", name);
		 
    if (fh = fopen(logF, "w"))
    {
      (void) fprintf(fh, "%ld\n", (long) getpid());
      fclose(fh);
      return 0;
    }
  }
  return (-1);
}


pid_t 
daemon_init(ttyf, name)
	int ttyf;
	char *name;
{
	char *tty;
	pid_t pid;
	int i;

	daemon(1, 1);				/* just fork() etc */
	
	if (ttyf) {
		int devnull = open("/dev/null", O_RDWR, 0);
		
		for (i = 0; i < MAX_FD; i++) {
			if (isatty(i) && ttyf == 1) {
				/* 
				 * On some systems it is necessary to
				 * actually close tty devices.  At
				 * this point though we can safely
				 * re-open them.  
				 */
				tty = ttyname(i);
				(void) close(i);
				if (open(tty, O_RDWR) != i) {
					if (i < 2)
						perror("open(/dev/tty)");
					ERRET;	/* NOT REACHED */
				}
			} else {
				(void) close(i);
				(void) dup(devnull, i);
			}
		}
		
		if (devnull > 2)
			(void) close(devnull);
	}

	pid = getpid();

	if (name)
		(void) LogMyPID(pid, name);

	return pid;
}


#ifdef MAIN
int 
main (int argc, char **argv)
{
  puts("calling daemon_init");
  fflush(stdout);
  printf("PID==%d\n", daemon_init(atoi(argv[1]), "daemon"));
  sleep(60);
  fprintf(stderr, "bye");
  exit(0);
}
#endif
#endif
