/* NAME:
 *	unfsd - user NFS daemon
 *
 * SYNOPSIS:
 *	unfsd [-avrdn][-f "exports"]
 *
 * DESCRIPTION:
 *	This is a unified mountd/nfsd derrived from the Linux NFS
 *	server.  This modified version can be run from inetd without
 *	the need for a portmapper, by using 'unfsc' on the client
 *	system.  Otherwise it behaves much like the original server.
 *
 *	Options:
 *
 *	-r	Re-export mounted filesystems.
 *
 *	-f "exports"
 *		Use "exports" rather than the default.
 *
 *	-a	Require strong authentication (one time password) from
 *		client before allowing mount to proceed.  This
 *		currently uses the authsrv from the TIS fwtk.
 *
 *	-d	Print lots of verbose debugging information.  Only
 *		needed when debugging the server.
 *
 *	-m	Register with portmapper.
 *
 *	-n	Allow non-root mount requests.
 *
 *	-v	Show version and exit.
 *
 *	Normally for both TCP and UDP transports, the server exits
 *	after an idle timeout.  For UDP this is normal.  For TCP the
 *	client 'unfsc' will automagically re-establish the connection
 *	when needed.
 *
 *	The above plan breaks if we run with '-a' and require a
 *	challenge/response before allowing a mount.  In this case the
 *	client attempts to unmount the filesystem and then exit.  To
 *	avoid this, the client ``pings'' us when it has been idle for
 *	a while which causes the server to maintain the session. 
 *	
 * SEE ALSO:
 *	exports(5), unfsc(8).
 *
 * AUTHOR:
 *	Simon J. Gerraty <sjg@zen.void.oz.au>
 *	
 */
/*
 *      @(#)Copyright (c) 1994, Simon J. Gerraty.
 *      
 *      This code is presented AS IS.  It comes with NO WARRANTY.
 *      Permission to use, modify and distribute this source code 
 *      is granted subject to the following conditions.
 *      1/ that the above copyright notice and this notice 
 *      are preserved in all copies and that due credit be given 
 *      to the author.  
 *      2/ that any changes to this code are clearly commented 
 *      as such so that the author does not get blamed for bugs 
 *      other than his own.
 *      
 *      Please send copies of changes and bug-fixes to:
 *      sjg@zen.void.oz.au
 */
#ifndef lint
static char RCSid[] = "$Id: unfsd.c,v 1.9 1995/05/13 07:23:48 sjg Exp $";
#endif
#include <stdio.h>
#include <syslog.h>

/*
 * DO NOT EDIT THIS UNLESS YOU _KNOW_ WHAT YOU ARE DOING...
 * It looks like it might have been generated by rpcgen and originally
 * some of it was... but it has been hacked mightily to run cleanly under
 * inetd and without use of portmap.
 */

#include "nfsd.h"
#include "unfsd.h"
#include "pathnames.h"

#ifdef DEBUG
#define RPC_SVC_FG
#endif

char *Myname = "unfsd";

void nfs_dispatch	__P((struct svc_req * rqstp, SVCXPRT * transp));
void mountprog_1	__P((struct svc_req * rqstp, SVCXPRT * transp));

/* not all RPC headers declare this */
SVCXPRT *svcfd_create	__P((int fd, unsigned int sendsize, unsigned int recvsize));

extern char *auth_file;
extern char argbuf[MNTPATHLEN + 1];

static int NeedAuth = 0;
int WantExit = 0;

#ifndef NO_FWTK
static struct timeval TIMEOUT = { 10, 0 };
static CLIENT *UnfsClp = NULL;
#endif


static void
cleanup()
{
	dprintf(1, "shutting down...");

#ifdef DO_UNFS_PROG
#ifndef NO_FWTK
	if (NeedAuth && UnfsClp != NULL) {
		char res = 0;

		if (clnt_call(UnfsClp, UNFS_QUIT,
			      xdr_void, &res,
			      xdr_void, &res,
			      TIMEOUT) == RPC_SUCCESS) {
			syslog(LOG_DEBUG, "told client to shutdown");
		}
	}
#endif
#endif
}

/*
 * For both TCP and UPD transports, we want to clear out if idle.
 * In both cases, client will re-establish connection if needed.
 * For TCP this is easier than trying to detect if the client is still
 * alive.
 */

static void
closedown()
{
	if (_rpcsvcdirty == 0) {
#if DO_UNFS_PROG
#ifndef NO_FWTK
		if (NeedAuth > 0 && UnfsClp != NULL) {
			/* test if client still there... */
			static ping_failures = 0;
			char res = 0;
			
			if (clnt_call(UnfsClp, UNFS_NULL,
				      xdr_void, &res,
				      xdr_void, &res,
				      TIMEOUT) == RPC_SUCCESS) {
				syslog(LOG_DEBUG, "pinged client");
				(void) alarm(_rpcsvc_closedown);
				ping_failures = 0;
				return;
			}
			syslog(LOG_WARNING, "no answer from client");
			if (ping_failures++ < 5)
				return;
			
			UnfsClp = NULL;
		}
#endif
#endif
		exit(0);
	}
	(void) alarm(_rpcsvc_closedown);
}


void
usage()
{
	(void) fprintf(stderr, "%s [-a][-v][-d][-r][-n][-f exports]", Myname);
	exit(1);
}

int
main(argc, argv)
	int argc;
	char **argv;
{
#ifndef BSD4_4
	extern	char *__progname;		/* warn*() want this */
#endif
	extern char version[];
	extern char *optarg;
	extern int getopt(), optind;
	register SVCXPRT *transp;
	int sock;
	int proto;
	struct sockaddr_in saddr, from;
	int asize = sizeof(saddr);
	int c, use_pmaper = 0;

	if ((Myname = strrchr(argv[0], '/')) != NULL)
		Myname++;
	else 
		Myname = argv[0];

#ifndef BSD4_4
	__progname = Myname;
#endif
	
	auth_file = _PATH_EXPORTS;
	
	while ((c = getopt(argc, argv, "avmndrf:")) != EOF) {
		switch (c) {
		case 'a':
			NeedAuth++;
			break;
		case 'm':
			use_pmaper = 1;
			break;
		case 'd':
			toggle_logging(0);
			break;
		case 'f':
			auth_file = optarg;
			break;
		case 'v':
			printf("%s\n", version);
			exit(0);
		case 'n':
			allow_non_root = 1;
			break;
		case 'r':
			re_export = 1;
			break;
		default:
			usage();
		}	
	}
	
	if (getsockname(0, (struct sockaddr *)&saddr, &asize) == 0) {
		int ssize = sizeof(int);

		if (saddr.sin_family != AF_INET)
			exit(1);
		if (getsockopt(0, SOL_SOCKET, SO_TYPE,
				(char *)&_rpcfdtype, &ssize) == -1)
			exit(1);
		sock = 0;
		_rpcpmstart = 1;
		proto = 0;
		asize = sizeof(from);
		if (getpeername(0, (struct sockaddr *)&from, &asize) == 0) {
			syslog(LOG_INFO, "call from %s",
			       ip_caller_id(&from, 0));
		}
	} else {
#ifndef RPC_SVC_FG
		daemon(0, 0);
#endif
		sock = RPC_ANYSOCK;
		if (use_pmaper) {
			(void) pmap_unset(MOUNTPROG, MOUNTVERS);
			(void) pmap_unset(NFS_PROGRAM, NFS_VERSION);
		}
	}

	if ((_rpcfdtype == 0) || (_rpcfdtype == SOCK_DGRAM)) {
		transp = svcudp_create(sock);
		if (transp == NULL) {
			_msgout("cannot create udp service.");
			exit(1);
		}
		if (!_rpcpmstart)
			proto = IPPROTO_UDP;
		if (!svc_register(transp, MOUNTPROG, MOUNTVERS, mountprog_1, proto)) {
			_msgout("unable to register (MOUNTPROG, MOUNTVERS, udp).");
			exit(1);
		}
		if (!svc_register(transp, NFS_PROGRAM, NFS_VERSION, nfs_dispatch, proto)) {
			_msgout("unable to register (NFS_PROGRAM, NFS_VERSION, tcp).");
			exit(1);
		}
		dprintf(1, "registered UDP services.");
	}

	if ((_rpcfdtype == 0) || (_rpcfdtype == SOCK_STREAM)) {
		if (sock == RPC_ANYSOCK)
			transp = svctcp_create(sock, 0, 0);
		else {
			transp = svcfd_create(sock, 0, 0); /* already connected */
			/*
			 * we need this to authenticate caller
			 * see rendezvous_request(svc_tcp.c)
			 */
			transp->xp_raddr = from;
			transp->xp_addrlen = asize;
		}
		if (transp == NULL) {
			_msgout("cannot create tcp service.");
			exit(1);
		}
		if (!_rpcpmstart)
			proto = IPPROTO_TCP;
		if (!svc_register(transp, MOUNTPROG, MOUNTVERS, mountprog_1, proto)) {
			_msgout("unable to register (MOUNTPROG, MOUNTVERS, tcp).");
			exit(1);
		}
		if (!svc_register(transp, NFS_PROGRAM, NFS_VERSION, nfs_dispatch, proto)) {
			_msgout("unable to register (NFS_PROGRAM, NFS_VERSION, tcp).");
			exit(1);
		}
		dprintf(1, "registered TCP services.");
	}

	if (transp == (SVCXPRT *)NULL) {
		_msgout("could not create a handle");
		exit(1);
	}

	atexit(cleanup);
	(void) signal(SIGALRM, closedown);
	(void) signal(SIGTERM, cleanup);

	if (_rpcpmstart) {
		_rpcsvc_closedown = _RPCSVC_CLOSEDOWN;
		(void) alarm(_rpcsvc_closedown);
	}

	openlog(Myname, LOG_PID, LOG_LOCAL1);

#ifndef NO_FWTK
	if (NeedAuth && _rpcpmstart && _rpcfdtype == SOCK_STREAM) {
		do_auth(0, Myname);
#if DO_UNFS_PROG
		/*
		 * Ok, now we need to be able to ping client, all
		 * using the same socket!
		 */
		UnfsClp = clnttcp_create(&from, 
					 UNFS_PROG, UNFS_VERS,
					 &sock, 0, 0);
		UnfsClp->cl_auth = authunix_create_default();
#endif
	} else
		NeedAuth = 0;			/* ignore it */
#endif
	/* Initialize the FH module. */
	fh_init();

	/* Initialize the AUTH module. */
	auth_init(auth_file);

	/* Enable the LOG toggle with a signal. */
	signal(SIGUSR1, toggle_logging);
	
	dprintf(1, "calling svc_run()...");
	svc_run();
	_msgout("svc_run returned");
	exit(1);
	/* NOTREACHED */
}

/* this just the mount dispatch stuff */

#ifndef HAVE_RPCGEN_C
#define mountproc_null_1_svc		mountproc_null_1
#define mountproc_mnt_1_svc		mountproc_mnt_1
#define mountproc_dump_1_svc		mountproc_dump_1
#define mountproc_umnt_1_svc		mountproc_umnt_1
#define mountproc_umntall_1_svc		mountproc_umntall_1
#define mountproc_export_1_svc		mountproc_export_1
#define mountproc_exportall_1_svc	mountproc_exportall_1
#endif

void
mountprog_1(rqstp, transp)
	struct svc_req *rqstp;
	SVCXPRT *transp;
{
	union {
		dirpath mountproc_mnt_1_arg;
		dirpath mountproc_umnt_1_arg;
	} argument;
	char *result;
	bool_t (*xdr_argument)(), (*xdr_result)();
	char *(*local)();

	alarm(_rpcsvc_closedown);		/* reset idle timeout */
	_rpcsvcdirty = 1;
	dprintf(1, "mountprog_1(%d)",
	       rqstp->rq_proc);
	
	switch (rqstp->rq_proc) {
	case MOUNTPROC_NULL:
		xdr_argument = xdr_void;
		xdr_result = xdr_void;
		local = (char *(*)()) mountproc_null_1_svc;
		break;

	case MOUNTPROC_MNT:
		xdr_argument = xdr_dirpath;
		xdr_result = xdr_fhstatus;
		local = (char *(*)()) mountproc_mnt_1_svc;
		break;

	case MOUNTPROC_UMNT:
		xdr_argument = xdr_dirpath;
		xdr_result = xdr_void;
		local = (char *(*)()) mountproc_umnt_1_svc;
		WantExit = 1;
#if DO_UNFS_PROG
#ifndef NO_FWTK
		UnfsClp = NULL;			/* don't ping anymore */
#endif
#endif
		alarm(10);			/* wake us up after we've replied */
		break;
#ifdef notyet
	case MOUNTPROC_DUMP:
		xdr_argument = xdr_void;
		xdr_result = xdr_mountlist;
		local = (char *(*)()) mountproc_dump_1_svc;
		break;

	case MOUNTPROC_UMNTALL:
		xdr_argument = xdr_void;
		xdr_result = xdr_void;
		local = (char *(*)()) mountproc_umntall_1_svc;
		break;

	case MOUNTPROC_EXPORT:
		xdr_argument = xdr_void;
		xdr_result = xdr_exports;
		local = (char *(*)()) mountproc_export_1_svc;
		break;

	case MOUNTPROC_EXPORTALL:
		xdr_argument = xdr_void;
		xdr_result = xdr_exports;
		local = (char *(*)()) mountproc_exportall_1_svc;
		break;
#endif
	default:
		svcerr_noproc(transp);
		_rpcsvcdirty = 0;
		return;
	}
	bzero((char *)&argument, sizeof(argument));
	if (!svc_getargs(transp, xdr_argument, &argument)) {
		svcerr_decode(transp);
		_rpcsvcdirty = 0;
		return;
	}
	result = (*local)(&argument, rqstp);
	if (result != NULL && !svc_sendreply(transp, xdr_result, result)) {
		svcerr_systemerr(transp);
	}
	if (!svc_freeargs(transp, xdr_argument, &argument)) {
		_msgout("unable to free arguments");
		exit(1);
	}
	_rpcsvcdirty = 0;
	return;
}
